import glob
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as np
import platform
import logging
logger = logging.getLogger(__name__)

__all__ = ['get_matlab', 'is_facet_srv']


def get_matlab(display=False, splash=False):
    system = platform.system()

    if is_facet_srv():
        matlab_base = 'fmatlab'
    elif system == 'Darwin':
        matlabs = glob.glob('/Applications/MATLAB_R*/bin/matlab')
        if np.size(matlabs) > 1:
            logger.warning('Multiple matlabs found!')
        matlab_base = matlabs[0]
    elif system == 'Linux':
        home = _os.environ['HOME']
        matlab_base = _os.path.join(home, 'Matlab/bin/./matlab')
       
    options = ''
    if not display:
        options = options + ' -nodisplay'

    if not splash:
        options = options + ' -nosplash'

    return matlab_base + options


def is_facet_srv():
    nodename = platform.node()
    if nodename[0:-2] == 'facet-srv':
        return True
    else:
        return False



import inspect as _ins
import os
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    import re as _np
else:
    import numpy as _np


__all__ = ['E200_Dat']


class E200_Dat(object):
    """
    Contains data. This is designed to prevent accidental incorrect correlation of data with UIDs, in case a request for data cannot find UIDs.

    **Should not be created directly, only accessed through** :func:`E200.E200_load_images`\ **.**
    """
    def __dir__(self):
        return [val[0] for val in _ins.getmembers(self)]

    def __init__(self, dat, uid, field):
        self._dat = _np.array([dat]).flatten()
        self._uid = _np.int64([uid]).flatten()
        self._field = field

    def __len__(self):
        return len(self._uid)

    @property
    def dat(self):
        """
        An array of data correlated to :attr:`E200.E200_Dat.UID`.
        """
        return self._dat

    def _get_uid(self):
        """
        An array of UIDs correlated to :attr:`E200.E200_Dat.dat`.
        """
        return self._uid
    uid = property(_get_uid)
    UID = property(_get_uid)

    @property
    def field(self):
        """
        Field data was loaded from.
        """
        return self._field



import sys as _sys
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    from PyQt4 import QtGui
    import scisalt.qt as mtqt

import configparser as _ConfigParser
import inspect as _inspect
import logging as _logging
import subprocess as _subprocess
import shlex as _shlex
from .get_matlab import get_matlab

logger = _logging.getLogger(__name__)

__all__    = ['get_remoteprefix', 'set_remoteprefix', 'choose_remoteprefix', '_get_configpath', '_get_directory', '_get_datapath']
def_prefix = '/Volumes/PWFA_4big'


def choose_remoteprefix(pathstart=def_prefix):
    app = mtqt.get_app()  # NOQA
    if not _os.path.isdir(pathstart):
        pathstart = _os.environ.get('HOME', _os.path.abspath(_os.sep))
    prefix = mtqt.getExistingDirectory(caption='Change prefix', directory=pathstart)
    if not _os.path.isdir(prefix):
        logger.critical('No directory selected: terminating.')
        raise IOError('No directory selected.')

    prefix = _test_prefix(prefix)
    set_remoteprefix(prefix)

    return prefix


def get_remoteprefix():
    app = mtqt.get_app()  # NOQA

    # =====================================
    # Default prefix
    # =====================================

    config_path = _get_configpath()

    # =====================================
    # Try to read the prefix
    # =====================================
    config = _ConfigParser.ConfigParser()
    try:
        config.read(config_path)
        prefix = config.get('Prefs', 'prefix')
    except:
        prefix = def_prefix
        set_remoteprefix(prefix)

    # =====================================
    # Test the prefix
    # =====================================
    prefix = _test_prefix(prefix)

    return prefix


def _get_datapath(prefix=None):
    if prefix is None:
        prefix = get_remoteprefix()
    datapath = _os.path.join(prefix, 'nas', 'nas-li20-pm00')
    return datapath


def _nas_in_path(prefix):
    datapath = _get_datapath(prefix)
    return _os.path.isdir(datapath)


def _test_prefix(prefix):
    while ( not _nas_in_path(prefix) ):
        title = 'Data location not found!'
        maintext = 'WARNING: Path to data doesn\'t exist!'
        infotext = 'Data not found at {}. Drive may not be mounted.'.format(prefix)
        buttons = _np.array([
            mtqt.Button('Try again.', QtGui.QMessageBox.AcceptRole, default=True),
            mtqt.Button('Locate folder containing /nas.', QtGui.QMessageBox.AcceptRole),
            mtqt.Button(QtGui.QMessageBox.Abort, escape=True),
            ])

        buttonbox = mtqt.ButtonMsg(title=title, maintext=maintext, infotext=infotext, buttons=buttons)
        clicked = buttonbox.clickedArray
        # Locate folder...
        if clicked[1]:
            prefix = choose_remoteprefix()
        elif clicked[2]:
            logger.critical('No valid directory chosen: terminating.')
            raise IOError('No valid directory chosen.')

    return prefix


def set_remoteprefix(prefix):
    config = _ConfigParser.ConfigParser()
    try:
        config.add_section('Prefs')
    except:
        pass

    config.set('Prefs', 'prefix', prefix)
    with open(_get_configpath(), 'wt') as configfile:
        config.write(configfile)

    matlab = get_matlab()
    command = '{matlab} -r "setpref(\'FACET_data\',\'prefix\',\'{prefix}\');exit;"'.format(matlab=matlab, prefix=prefix)

    logger.info('Setting remote prefix in Matlab: {}'.format(prefix))

    fnull = open(_os.devnull, 'w')
    _subprocess.call(_shlex.split(command), stdout=fnull, stderr=_subprocess.STDOUT)

    logger.info('Remote prefix now: {}'.format(prefix))


def _get_configpath():
    this_path   = _inspect.stack()[0][1]
    this_dir    = _os.path.dirname(this_path)
    config_path = _os.path.join(this_dir, 'FACET_data.cfg')

    return config_path


def _get_directory():
    window = QtGui.QFileDialog.getExistingDirectory(directory='/')
    return str(window)



import E200
import numpy as np
import scipy.optimize as spopt
import matplotlib.pyplot as plt
import scisalt as mt
__all__ = ['eaxis', 'eaxis_ELANEX', 'yaxis_ELANEX', 'yanalytic', 'E_no_eta', 'y_no_eta']

import logging
loggerlevel = logging.DEBUG
logger      = logging.getLogger(__name__)


def eaxis(y, uid, camname, hdf5_data, E0=20.35, etay=0, etapy=0):
    logger.log(level=loggerlevel, msg='Getting energy axis...')

    # eaxis     = E200.eaxis(camname=camname, y=y, res=res, E0=20.35, etay=0, etapy=0, ymotor=ymotor)
    imgstr = hdf5_data['raw']['images'][str(camname)]
    res    = np.float64(imgstr['RESOLUTION'][0, 0])
    res    = res*np.float64(1.0e-6)

    logger.log(level=loggerlevel, msg='Camera detected: {}'.format(camname))
    if camname == 'ELANEX':
        ymotor = hdf5_data['raw']['scalars']['XPS_LI20_DWFA_M5']['dat']
        ymotor = mt.derefdataset(ymotor, hdf5_data.file)
        ymotor = ymotor[0]*1e-3
        logger.log(level=loggerlevel, msg='Original ymotor is: {}'.format(ymotor))

        raw_rf     = hdf5_data['raw']
        scalars_rf = raw_rf['scalars']
        setQS_str  = scalars_rf['step_value']
        setQS_dat  = E200.E200_api_getdat(setQS_str, uid).dat[0]
        setQS      = mt.hardcode.setQS(setQS_dat)

        logger.log(level=loggerlevel, msg='Eaxis''s setQS is: {}'.format(setQS_dat))

        ymotor = setQS.elanex_y_motor()*1e-3
        logger.log(level=loggerlevel, msg='Reconstructed ymotor is: {ymotor}'.format(ymotor=ymotor))

        return eaxis_ELANEX(y=y, res=res, etay=etay, etapy=etapy, ymotor=ymotor)

    elif camname == 'CMOS_FAR':
        return eaxis_CMOS_far(y=y, res=res, E0=E0, etay=etay, etapy=etapy)

    else:
        msg = 'No energy axis available for camera: {}'.format(camname)
        logger.log(level=loggerlevel, msg=msg)
        raise NotImplementedError(msg)


def eaxis_ELANEX(y, res, etay=None, etapy=None, ypinch=None, img=None, ymotor=None):
    ymotor = np.float64(ymotor)
    y      = y+ymotor/res

    #  y_motor_calibrated = np.float64(-1e-3)
    #  y_pinch_calibrated = np.float64(130)
    #  y_pixel_size       = np.float64(4.65e-6)
    #  E0 = 20.35

    #  y_motor_calibrated = np.float64(-0.00677574370709)
    #  y_pinch_calibrated = np.float64(211)
    #  y_pixel_size       = np.float64(8.9185e-6)
    E0 = 23.737805394397343771

    # ypinch = y_pinch_calibrated + y_motor_calibrated/y_pixel_size

    # E0=20.35 observed at 130px, motor position -1mm
    #  E0=20.35

    theta  = np.float64(6e-3)
    Lmag   = np.float64(2*4.889500000E-01)
    Ldrift = np.float64(8.792573)

    logger.log(level=loggerlevel, msg='ypinch is: {}'.format(ypinch))
    logger.log(level=loggerlevel, msg='ymotor is: {}'.format(ymotor))

    out = E_no_eta(y, ypinch, res, Ldrift, Lmag, E0, theta)
    return out


def yaxis_ELANEX(E, res, E0=None, etay=None, etapy=None, ypinch=None, img=None, ymotor=None):
    # y=0
    def merit_fcn(y, res, E0=None, etay=None, etapy=None, ypinch=None, img=None, ymotor=None):
        val = eaxis_ELANEX(y, res, E0, etay, etapy, ypinch, img, ymotor)
        print('===========')
        print(y)
        print(val)
        print(E)
        out = (E-val)**2
        print(out)
        return out

    etay   = 0
    etapy  = 0
    ypinch = 0
    img    = 0

    # args = np.array([res, E0, etay, etapy, ypinch, img, ymotor])
    args = (res, E0, etay, etapy, ypinch, img, ymotor)

    # print(merit_fcn(0, res, E0, etay, etapy, ypinch, img, ymotor))
    # print(args)
    outval = spopt.minimize(merit_fcn, x0=np.array([-4000]), args=args, tol=1e-7)
    print(outval)
    return outval.x[0]


def eaxis_CMOS_far(y, res, E0=None, etay=None, etapy=None, img=None):
    # The axis is flipped.  Since the offset is arbitrary and
    # calibrated for below, I can use an arbitrary offset here
    # in order to flip the axis. I've decided on a random value of 4000
    # since I'm pretty sure none of our photos have a resolution
    # greater than that, and I'm trying to error out on a negative
    # value.
    ypinch = 1660
    y_flip_offset = max(y) + 10
    y = y_flip_offset - y
    ypinch = y_flip_offset - ypinch
    if min(y) < 0:
        raise ValueError('y<0 indicates my arbitrary offset is bad. Change source code!')
    if etay is None:
        etay = input('Dispersion in y in mm (eta_y)? ')
        etay = etay * 1e-3
        print('Dispersion entered is {}'.format(etay))

    if etapy is None:
        etapy = input('Dispersion-prime in y in mrad (eta''_y)? ')
        etapy = etapy * 1e-3
        print('Dispersion-prime entered is {}'.format(etapy))

    if img is not None:
        plt.imshow(img)
        plt.show()
    if ypinch is None:
        ypinch = input('Location of zero energy in y in pixels? ')

    # E0 = 20.35
    if E0 is None:
        E0 = input('Zero energy in GeV? ')
    theta = 6e-3

    Lmag   = 2*4.889500000E-01
    Ldrift = 8.792573 + 0.8198
    # out  = np.zeros(y.shape[0])
    # for i, yval in enumerate(y):
    #     args=np.array([yval, ypinch, res, E0, theta, Ldrift, Lmag, etay, etapy])
    #     outval=spopt.minimize(merit_fcn, x0=np.array([20]), args=args)
    #     out[i]=outval.x[0]
    #
    out = E_no_eta(y, ypinch, res, Ldrift, Lmag, E0, theta)
    return out

# def merit_fcn(E, ypx, ypinch, res, E0, theta, Ldrift, Lmag, eta0=np.float64(0), etap0=np.float64(0)):
#     E=E[0]
#     yoffset=yanalytic(E0, E0, theta, Ldrift, Lmag, eta0, etap0) - ypinch*res
#     y = ypx*res + yoffset
#     yana = yanalytic(E, E0, theta, Ldrift, Lmag, eta0, etap0)
#
#     # print('=====================')
#     # print(E)
#     # print(y)
#     # print(yana)
#     # print(np.power(y-yana, 2)*1e14)
#     # print('=====================')
#     return np.power(y-yana, 2)*1e14


def yanalytic(E, E0, theta, Ldrift, Lmag, eta0, etap0):
    logger.critical('hello')
    return y_no_eta(E, E0, theta, Ldrift, Lmag) + (E/E0 - 1)*(eta0+etap0*(Ldrift+Lmag))


def y_no_eta(E, E0, theta, Ldrift, Lmag):
    logger.critical('intermediate')
    output = Ldrift/np.sqrt(np.power(E/(E0*np.sin(theta)), 2)-1) + (E*Lmag)/(E0*np.sin(theta)) * (1 - np.sqrt(1-np.power(E0*np.sin(theta)/E, 2)))
    logger.critical('{}'.format(output))
    return output


# def my_E_no_eta(y, E0, theta, Ldrift, Lmag):

def E_no_eta(ypx, ypinch, res, Ldrift, Lmag, E0, theta, dataset_num=None):
    # Determine pinch location
    #  ypinch = y_pinch_calibrated + y_motor_calibrated/y_pixel_size
    #  E0 = 23.737805394397343771
    E0 = 20.35

    eta_0_meters = None
    if dataset_num == 13437 or dataset_num == 13438:
        y0    = np.float64(1589)    # pixel position of E0 (20.35 GeV).
        eta_0 = np.float64(949.72)  # nominal dipole dispersion in pixel, corresponding to 59.5 mm.
        
    elif dataset_num == 13448 or dataset_num == 13449:
        y0    = np.float64(1605.5) - np.float64(0.7923)*(E0+QS)  # y0 is adjusted to account for QS dispersion.
        eta_0 = np.float64(949.72) + np.float64(0.7923)*(E0+QS)  # added QS dispersion of 0.7923 pix per QS GeV.
        #  eta_0 = np.float64(59.5e-3) + np.float64(0.7923)*(E0+QS)
            
    elif dataset_num == 13450:
        y0    = np.float64(1655)   - np.float64(3.321)*(E0+QS)  # y0 is adjusted to account for QS dispersion.
        eta_0 = np.float64(949.72) + np.float64(3.321)*(E0+QS)  # added QS dispersion of 3.321 pix per QS GeV.
        
    elif dataset_num == 13537:
        y0    = np.float64(1576)   + np.float(0.5193)*(20.35+QS)  # y0 is adjusted to account for QS dispersion.
        eta_0 = np.float64(949.72) - np.float(0.5193)*(20.35+QS)  # added QS dispersion of -0.5193 pix per QS GeV.
    else:
        y0   = np.float64(1589)
        eta_0_meters = (Ldrift+Lmag/np.float64(2))*theta

    # ===========================
    # Sebastien's code
    # ===========================
    z_B5D36    = np.float64(2005.65085 )  # middle of dipole magnet
    z_ELANEX   = np.float64(2015.22    )  # linac z location of ELANEX phosphor screen in meter
    z_CFAR     = np.float64(2016.04    )  # linac z location of Cherenkov Far gap in meter
    cal_ELANEX = np.float64(8.9185     )  # ELANEX camera calibration in um/pixel
    cal_CFAR   = np.float64(62.65      )  # CMOS FAR camera calibration in um/pixel
    # y0 = 259 (when QS=0) at ELANEX corresponds to y0 = 1589 on CMOS FAR.
    y0    = np.float64(259) + (cal_CFAR/cal_ELANEX) * (y0-np.float64(1589))
    ypinch = y0

    # ===========================
    # vvv Moved Code
    # ===========================

    yoffset = yanalytic(E0, E0, theta, Ldrift, Lmag, eta0=0, etap0=0) - ypinch*res

    y = ypx*res + yoffset

    if type(y) != np.ndarray:
        y = np.array([y])

    # ===========================
    # ^^^^ Moved Code
    # ===========================

    eta_0 = (cal_CFAR/cal_ELANEX) * (z_ELANEX-z_B5D36) / (z_CFAR-z_B5D36) * eta_0

    if eta_0_meters is None:
        eta_0_meters = eta_0*cal_CFAR*np.float64(1e-6)

    logger.debug('Dispersion is: {}'.format(eta_0_meters))

    approx = E0*eta_0_meters / y

    def merit(Eguess, yval):
        yfromE = y_no_eta(Eguess, E0, theta, Ldrift, Lmag)
        return np.power(yfromE-yval, 2)
    #  #
    results = np.zeros(y.size, dtype=np.float64)
    for i, yval in enumerate(y):
        guess      = E0*(np.float64(2)*Ldrift+Lmag)*theta / (np.float64(2)*yval)
        results[i] = spopt.fmin(merit, guess, args=(yval, ), disp=False)
    
    return results, approx



import os
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    import re as _np
    import re as _h5
else:
    import numpy as _np
    import h5py as _h5
import uuid as _uuid
import warnings

import logging
logger = logging.getLogger(__name__)
#  loggerlevel = logging.DEBUG
loggerlevel = 9


def E200_api_updateUID(group, UID, value, verbose=False):
    if _np.size(UID) > 1:
        # ======================================
        # Run through list of UID's, processing
        # individually
        # ======================================
        logger.info('Multiple UIDs entered, updating recursively')
        
        # ======================================
        # Check for duplicate UIDs and sort
        # ======================================
        UID_sorted, inds = _np.unique(UID, return_index=True)
        value_sorted    = value[inds]
        if _np.size(UID_sorted) != _np.size(UID):
            raise ValueError('UIDs aren''t unique: {}'.format(UID))
            
        for i, uid in enumerate(UID_sorted):
            E200_api_updateUID(group, value=value_sorted[i], UID=uid, verbose=verbose)
            
    else:
        # ref_dtype = _h5.special_dtype(ref=_h5.Reference)
        
        # ======================================
        # Get basic dsets
        # ======================================
        uid_dset  = group['UID']
        dat_dset  = group['dat']
        refs_dset = group['refs']
        
        uids      = uid_dset.value
        # dat_refs  = dat_dset.value
        
        logger.log(level=loggerlevel, msg='UID to update: {}, value: {}'.format(UID, value))
        
        # ======================================
        # Validate it's a numpy array
        # ======================================
        value = _np.array(value)
        
        # ======================================
        # Find matches
        # ======================================
        uid_match = _np.in1d(uids, UID)
        uid_match_total = _np.sum(uid_match)
        uid_match_inds = _np.where(uid_match)[0]
        
        if uid_match_total == 0:
            # ======================================
            # Add new UID and value
            # ======================================
            logger.log(level=loggerlevel, msg='Adding new UID...')
            
            uuid_val = str(_uuid.uuid1())
            uuid_ref = refs_dset.create_dataset(name=uuid_val, data=value)
            
            uids_shape = uid_dset.shape
            
            uid_dset.resize((uids_shape[0]+1, ))
            dat_dset.resize((uids_shape[0]+1, ))
            
            uid_dset[-1] = UID
            dat_dset[-1] = uuid_ref.ref
            
        elif uid_match_total == 1:
            # ======================================
            # Delete old UID dataset
            # ======================================
            logger.log(level=loggerlevel, msg='Replacing existing UID...')
            ind = uid_match_inds[0]
            ref = dat_dset.value[ind]
            ref_name = group.file[ref].name
            split = ref_name.split('/')
            ref_uuid = split[-1]
            del refs_dset[ref_uuid]
            
            # ======================================
            # Create new UUID dataset
            # ======================================
            uuid_val = str(_uuid.uuid1())
            uuid_ref = refs_dset.create_dataset(name=uuid_val, data=value)
            
            # ======================================
            # Replace reference
            # ======================================
            dat_dset[ind] = uuid_ref.ref
        elif uid_match_total > 1:
                # ======================================
                # Too many matches, corruption?
                # ======================================
                raise LookupError('Multiple unique IDs exist. ({} uids with values: {})'.format(uid_match_total, uids[uid_match]))
                
        # ======================================
        # Write changes to file
        # ======================================
        group.file.flush()



from .E200_dataset2str import *     # NOQA
from .classes import *              # NOQA
from .get_valid_filename import *   # NOQA
from .get_matlab import get_matlab
import logging
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import h5py as _h5
import shlex as _shlex
import subprocess as _subprocess
import tempfile as _tempfile
loggerlevel = logging.DEBUG
logger      = logging.getLogger(__name__)

__all__ = ['E200_load_data']


def E200_load_data(filename, savefile=None, matlab_verbose=False):
    """
    Loads  dataset file where *filename* is a :code:`str` of the relative location of the file (i.e. ``nas/nas-li20-pm00/E200/2015/20150602/E200_17712``). If specified, saves the intermediate h5py file to *savefile*. *matlab_verbose* determines whether Matlab's output while loading the file is printed to the terminal.

    Returns an instance of :class:`E200.Data`.

    *Note: this function calls Matlab code; it is not surprising to see Matlab open in the terminal.*
    """
    logger.log(level = loggerlevel, msg = 'Input is: filename={}'.format(filename))

    # ======================================
    # Create temporary directory
    # ======================================
    if savefile is None:
        with _tempfile.TemporaryDirectory() as tempdir:
            tempfilename = 'temp.h5'
            temppath     = _os.path.join(tempdir, tempfilename)
            return _process_file(filename=filename, temppath=temppath, matlab_verbose=matlab_verbose)
    else:
        return _process_file(filename=filename, temppath=savefile, matlab_verbose=matlab_verbose)


def _process_file(filename, temppath, matlab_verbose):
    # ======================================
    # Have matlab process file
    # ======================================
    logger.log(level=loggerlevel, msg='Processed file not found, calling matlab to process file.')
    pwd           = _os.getcwd()
    matlab        = get_matlab()
    curdir        = _os.path.dirname(_os.path.realpath(__file__))
    command       = '{matlab} -r "addpath(fullfile(\'{curdir}\',\'matlab\'));convert_mat_file(\'{filename}\',\'{outfile}\');exit;"'.format(matlab=matlab, curdir=curdir, pwd=pwd, filename=filename, outfile=temppath)
    command_split = _shlex.split(command)
    
    logger.log(level=loggerlevel, msg='Command given is: {}'.format(command))
    if matlab_verbose:
        fnull = open(_os.devnull, 'w')
        _subprocess.call(command_split, stdout=fnull)
    else:
        _subprocess.call(command_split)

    # ======================================
    # Load file processed by matlab
    # ======================================
    logger.log(level=loggerlevel, msg='Loading processed file')
    f = _h5.File(temppath, 'r', driver='core', backing_store=False)

    output = Data(read_file = f, filename=filename)

    return output


def _load_data(wf):
    processed = wf.create_group('/data/processed')
    groups    = ['arrays', 'scalars', 'images', 'vectors']
    for val in groups:
        processed.create_group(val)

    wf.flush()


class datalevel(object):
    pass


def recursivePopulate(h5data, objData):
    for i, val in enumerate(h5data):
        if type(h5data[val]) == _h5._hl.group.Group:
            setattr(objData, val, datalevel())
            recursivePopulate(h5data[val], getattr(objData, val))
        else:
            setattr(objData, val, h5data[val])



from .E200_load_data import E200_load_data
from .get_remoteprefix import get_remoteprefix
import glob
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import PyQt4.QtGui as QtGui
    import scisalt as pt
import re
import logging
logger = logging.getLogger(__name__)


def E200_load_data_gui(experiment=None, verbose=False):
    """
    Allows user to select a dataset file to load via a GUI file picker.

    If *experiment* (type :code:`str`) is specified, the GUI defaults to the last saved dataset from *experiment*. Typical options are:

    * :code:`'E200'`
    * :code:`'E217'`

    Returns an instance of :class:`E200.Data`.

    *Note: this function calls Matlab code; it is not surprising to see Matlab open in the terminal.*
    """
    if experiment is not None:
        recent = 'nas/nas-li20-pm00/{}'.format(experiment)
    else:
        recent = 'nas/nas-li20-pm00/E*'

    # ======================================
    # Get most recent folder
    # ======================================
    pref = get_remoteprefix()
    temppath = _os.path.join(pref, recent)
    temppath = max(glob.glob(_os.path.join(temppath, '*')), key=_os.path.getmtime)
    temppath = max(glob.glob(_os.path.join(temppath, '*')), key=_os.path.getmtime)
    temppath = max(glob.glob(_os.path.join(temppath, '*')), key=_os.path.getmtime)

    # ======================================
    # User selects file
    # ======================================
    app = pt.qt.get_app()  # NOQA
    loadfile = QtGui.QFileDialog.getOpenFileName(directory=temppath, filter='*.mat')
    if loadfile == '':
        input('No file chosen, press enter to close...')
        return
    loadfile  = loadfile[1:]
    p         = re.compile('nas/nas.*')
    loadmatch = p.search(loadfile)
    loadfile  = loadmatch.group()
    # loadfile  = 'nas/nas-li20-pm00/E217/2015/20150504/E217_16808/E217_16808.mat'

    logstring = 'File to load is: {}'.format(loadfile)
    if verbose:
        print('In order to load file:\ndata = E200.E200_load_data(\'{}\')'.format(loadfile))

    logger.info(logstring)

    return E200_load_data(loadfile)



import logging
loggerlevel = 9
loggerlevel = logging.DEBUG
logger      = logging.getLogger(__name__)

from .get_remoteprefix import *  # NOQA
from glob import glob as _glob
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    import scisalt as _mt
    import scisalt.qt as _mtqt
    from PyQt4 import QtGui as _QtGui
import os.path as _path
import re as _re
import warnings as _warnings


class Filename(object):
    def __init__(self, path, local=False):
        logger.log(level=loggerlevel, msg='Entering Filename class')
        self._local = local
        self.pathstr = path

    @property
    def experiment(self):
        match = _re.search('(?<=/)E\d{3}(?=/)', self.pathstr)
        experiment = match.group()
        return experiment

    # Return pathstr and set pathstr
    def _get_pathstr(self):
        return self._pathstr

    def _set_pathstr(self, value):
        # Reprocess pathstr
        self._pathstr          = value
        out                    = get_valid_filename(value, experiment=self.experiment, local=self._local)
        self._dir_beg          = out[0]
        self._dir_mid          = out[1]
        self._filename         = out[2]
        self._data_source_type = out[3]
    pathstr = property(_get_pathstr, _set_pathstr)

    # Retrieve dir_beg
    def _get_dir_beg(self):
        return self._dir_beg
    dir_beg = property(_get_dir_beg)

    # Retrieve dir_mid
    def _get_dir_mid(self):
        return self._dir_mid
    dir_mid = property(_get_dir_mid)

    # Retrieve filename
    def _get_filename(self):
        return self._filename
    filename = property(_get_filename)

    # Retrieve valid path
    def _get_valid_path(self):
        return '/'.join([self.dir_beg, self.dir_mid, self.filename])
    valid_path = property(_get_valid_path)

    # Retrieve data_source_type
    def _get_data_source_type(self):
        return self._data_source_type
    data_source_type = property(_get_data_source_type)

    def _get_processed_dir(self):
        processed_dir = _path.join(self.dir_beg, self.dir_mid)
        if not self._local:
            processed_dir = _re.sub('nas/nas-li20-pm0.', 'processed_data', processed_dir)
        return processed_dir
    processed_dir = property(_get_processed_dir)
    py_processed_dir = property(_get_processed_dir)

    def _get_processed_filename(self):
        fileparts = _path.splitext(self.filename)
        final_filename = '{}_processed{}'.format(fileparts[0], fileparts[1])
        return final_filename
    processed_filename = property(_get_processed_filename)

    def _get_processed_path(self):
        return _path.join(self.processed_dir, self.processed_filename)
    processed_path = property(_get_processed_path)

    def _get_py_processed_filename(self):
        return _re.sub('processed', 'py_processed', self.processed_filename)
    py_processed_filename = property(_get_py_processed_filename)
    
    def _get_py_processed_path(self):
        return _path.join(self.py_processed_dir, self.py_processed_filename)
    py_processed_path = property(_get_py_processed_path)


def get_valid_filename(pathstr, experiment, verbose=False, local=False):
    # File doesn't exist
    # This case must terminate in us trying again.
    logger.log(level=loggerlevel, msg='Input is: pathstr={}'.format(pathstr))

    # ======================================
    # Return info if the file is local
    # ======================================
    if local:
        if not _path.exists(pathstr):
            _warnings.warn('File not found: {}'.format(pathstr))
            #  raise IOError('File not found: {}'.format(pathstr))
        pathstr = _path.abspath(pathstr)
        dir_beg = ''
        dir_mid = _path.dirname(pathstr)
        filename = _path.basename(pathstr)
        data_source_type = '2014'
        output = (dir_beg, dir_mid, filename, data_source_type)
        return output

    # ======================================
    # Return info if the file is a nas file
    # ======================================
    if not _path.exists(pathstr):

        logger.log(level=loggerlevel, msg='Path does not exist: {}'.format(pathstr))
        # If file didn't exist, maybe it needs a prefix
        # Append prefix
        prefix = get_remoteprefix()
        pathstr = _path.join(prefix, pathstr)
        # print(prefix)

        # If it didn't work, we need to find the file somehow
        # The prefix may be wrong, the file may not be where we thought,
        # or the drive might not have been mounted and we can mount and
        # try again.
        if not _path.exists(pathstr):
            logger.log(level=loggerlevel, msg='Adding prefix didn''t help, new pathstr={}'.format(pathstr))

            title    = 'File not found!'
            maintext = 'File doesn''t exist:\n\n{}'.format(pathstr)
            subtext  = '(External drive may not be mounted.)'
            buttons  = _np.array([
                _mtqt.Button('Try again', _QtGui.QMessageBox.AcceptRole),
                _mtqt.Button('Change prefix', _QtGui.QMessageBox.ActionRole, default=True),
                _mtqt.Button('Locate file', _QtGui.QMessageBox.ActionRole),
                _QtGui.QMessageBox.Abort
                ])

            buttonbox = _mtqt.ButtonMsg(maintext, title=title, infotext=subtext, buttons=buttons)
            clicked   = buttonbox.clickedArray

            # Change prefix
            if clicked[0]:
                logger.log(level=loggerlevel, msg='Trying again')
            elif clicked[1]:
                logger.log(level=loggerlevel, msg='Changing prefix')

                prefix = _mt.E200.choose_remoteprefix(pathstart=prefix)
            # Locate file
            elif clicked[2]:
                logger.log(level=loggerlevel, msg='Locating file')

                # Try to open the path by chopping of the end
                while not _path.isdir(pathstr):
                    lastpath = pathstr
                    pathstr  = _path.dirname(pathstr)
                    # If you've chopped all you can chop, look in the base data folder
                    if lastpath == pathstr:
                        pathstr = _get_datapath()
                        break
                pathstr = _mtqt.getOpenFileName(directory=pathstr, caption='Locate file', filter='Matlab Files (*.mat)')
            elif clicked[3]:
                logger.log(level=loggerlevel, msg='Aborting')

                raise IOError('Aborting')
            else:
                print('Thar be dragons!')
        logger.log(level=loggerlevel, msg='Trying new pathstr={}'.format(pathstr))

        return get_valid_filename(pathstr)
    elif _path.isdir(pathstr):
        logger.log(level=loggerlevel, msg='Path is a directory')

        patterns = ['*scan_info.mat', '*filenames.mat', '{}_*.mat'.format(experiment)]
        files    = _np.array([])
        for i, val in enumerate(patterns):
            gfiles = _glob(_path.join(pathstr, val))
            files = _np.append(files, gfiles)
        logger.log(level=loggerlevel, msg='Files match patterns\npatterns: {}\nfiles: {}'.format(patterns, files))

        if files.size == 1:
            logger.log(level=loggerlevel, msg='Found one file.')

            pathstr = files[0]
        elif files.size == 0:
            logger.log(level=loggerlevel, msg='Found no files.')

            title    = 'File Not Found'
            maintext = 'No valid file found in folder.'
            subtext  = 'Locate manually?'
            buttons  = _np.array([
                _QtGui.QMessageBox.Ok,
                _QtGui.QMessageBox.Abort
                ], dtype=object)

            buttonbox = _mtqt.ButtonMsg(maintext, title=title, infotext=subtext, buttons=buttons)
            clicked   = buttonbox.clickedArray

            if clicked[0]:
                logger.log(level=loggerlevel, msg='Trying to locate manually.')

                pathstr = _mtqt.getOpenFileName(directory=pathstr, caption='Locate file', filter='Matlab Files (*.mat)')
            elif clicked[1]:
                logger.log(level=loggerlevel, msg='Aborting.')

                raise IOError('Aborting')
        logger.log(level=loggerlevel, msg='Trying new pathstr={}'.format(pathstr))

        return get_valid_filename(pathstr)

    elif _path.isfile(pathstr):
        logger.log(level=loggerlevel, msg='Path is a file.')

        # Check for string endings
        strends = _np.array(['scan_info.mat$', 'filenames.mat$', '{}_[0-9]*\.mat'.format(experiment)])
        strmatch = _np.empty(strends.size, dtype=object)
        for i, val in enumerate(strends):
            strmatch[i] = _re.search(val, pathstr)

        # True for matches
        strmatch_bool = _np.not_equal(strmatch, None)

        # Warn if there are no matches
        if not _np.any(strmatch_bool):
            logger.log(level=loggerlevel, msg='File doesn''t match patterns={}'.format(strends))

            _warnings.warn('Neither a 2014+ data file, scan_info.mat file, nor a filenames.mat file: {}'.format(pathstr))

        if strmatch_bool[2]:
            data_source_type = '2014'
        else:
            data_source_type = '2013'

        logger.log(level=loggerlevel, msg='data_source_type is {}'.format(data_source_type))

        # Check that this is a data file
        # Must have /nas/nas-li20-pm01
        match = _re.search('/nas/nas-li20-pm0', pathstr)
        if match is None:
            raise IOError('Does not point to a file with /nas/nas-li20-pm01 or /nas/nas-li20-pm00 in its path.')

        # Get the directories
        dirstr   = _path.dirname(pathstr)
        filename = _path.basename(pathstr)
        
        dir_beg = dirstr[:match.start()]
        dir_mid = dirstr[match.start()+1:]

        output = (dir_beg, dir_mid, filename, data_source_type)
        logger.log(level=loggerlevel, msg='Output is: {}'.format(output))

        return output



# Author: Joel Frederico <joelfred@slac.stanford.edu>.
__all__ = [
    'E200_Image_Iter',
    'E200_api_getUID',
    'E200_api_getdat',
    'E200_api_updateUID',
    'E200_create_data',
    'E200_dataset2str',
    'E200_load_data',
    'E200_load_data_gui',
    'E200_load_images',
    'classes'
    ]
__version__ = '1.3.1'
from .setQS import *                                 # noqa
from .E200_Image_Iter import *                       # noqa
from .E200_load_data import *                        # noqa
from .E200_load_data_gui import E200_load_data_gui   # noqa
from .E200_load_images import E200_load_images       # noqa
from .E200_create_data import *                      # noqa
from .E200_api_getUID import E200_api_getUID         # noqa
from .E200_api_getdat import *                       # noqa
from .E200_api_updateUID import *                    # noqa
from .E200_dataset2str import *                      # noqa
# from .eaxis import *                                 # noqa
# from .eaxis_class import *                           # noqa
# from eaxis import eaxis
# from eaxis import eaxis_ELANEX
# from eaxis import yaxis_ELANEX
# from eaxis import yanalytic
# from eaxis import E_no_eta
# from eaxis import y_no_eta

from .get_remoteprefix import *                      # noqa
from .get_valid_filename import *                    # noqa
from .classes import *                               # noqa



import E200
import scisalt as mt
import numpy as np

__all__ = ['Energy_Axis']

import logging
loggerlevel = logging.DEBUG
logger      = logging.getLogger(__name__)


class Energy_Axis(object):
    """Object that sets up energy function to compute energy axis"""
    # ===================================
    # Initialize class
    # ===================================
    def __init__(self, camname, hdf5_data, uid):
        """Initialize input parameters"""
        self.camname    = camname
        self._hdf5_data = hdf5_data
        self._uid       = uid
        self._reset_class()

    # ===================================
    # Reset hidden data
    # ===================================
    def _reset_class(self):
        """Resets hidden data in the class that all calculations are based off of"""
        self._dataset_num = None
        self._ypinch      = None
        self._setQS       = None

    def _get_hdf5_data(self):
        return self._hdf5_data

    def _set_hdf5_data(self, value):
        self._hdf5_data = value
        self._reset_class()
    hdf5_data = property(_get_hdf5_data, _set_hdf5_data)

    # ===================================
    # Get dataset number
    # ===================================
    def _get_dataset_num(self):
        """Get dataset number from first motor UID"""
        if self._dataset_num is None:
            dataset_num       = np.int64(self._hdf5_data['raw']['scalars']['XPS_LI20_DWFA_M5']['UID'].value[0][0])
            dataset_num       = str(dataset_num)
            dataset_num = dataset_num[0:5]
        return dataset_num
    dataset_num = property(_get_dataset_num)

    # ===================================
    # Get camera resolution
    # ===================================
    def _get_res(self):
        """Camera resolution"""
        imgstr = self._hdf5_data['raw']['images'][str(self.camname)]
        res    = np.float64(imgstr['RESOLUTION'][0, 0])
        res    = res*np.float64(1.0e-6)
        return res
    res = property(_get_res)

    # ===================================
    # Get ymotor for ELANEX from file
    # (not saved properly for all steps)
    # ===================================
    def _get_ymotor_orig(self):
        """Motor position from file"""
        #  if self.camname=='ELANEX':
        ymotor = hdf5_data['raw']['scalars']['XPS_LI20_DWFA_M5']['dat']
        ymotor = mt.derefdataset(ymotor, hdf5_data.file)
        ymotor = ymotor[0]*1e-3
        logger.log(level=loggerlevel, msg='Original ymotor is: {}'.format(ymotor))
        return ymotor
    ymotor_orig = property(_get_ymotor_orig)

    # ===================================
    # Get setQS equivalent
    # ===================================
    def _get_setQS(self):
        if self._setQS is None:
            '''Reconstruct quadrupole functions because motor values aren't recorded properly to file'''
            raw_rf     = self.hdf5_data['raw']
            scalars_rf = raw_rf['scalars']
            setQS_str  = scalars_rf['step_value']
            setQS_dat  = E200.E200_api_getdat(setQS_str, self._uid).dat[0]
            #  setQS_dat=0
            self._setQS      = mt.hardcode.setQS(setQS_dat)

            logger.log(level=loggerlevel, msg='Eaxis''s setQS is: {}'.format(setQS_dat))

        return self._setQS
    setQS = property(_get_setQS)

    # ===================================
    # Get ymotor from setQS
    # ===================================
    def _get_ymotor(self):
        ymotor = self.setQS.elanex_y_motor()*1e-3
        logger.log(level=loggerlevel, msg='Reconstructed ymotor is: {ymotor}'.format(ymotor=ymotor))
        return ymotor
    ymotor = property(_get_ymotor)

    # ===================================
    # Elements for calculating dispersion
    # ===================================
    _theta  = np.float64(6e-3)
    _Lmag   = np.float64(2*4.889500000E-01)

    # ===================================
    # Future-proofing, in case we want
    # to calculate dynamically
    # ===================================
    def _get_theta(self):
        return self._theta
    theta = property(_get_theta)

    def _get_Lmag(self):
        return self._Lmag
    Lmag = property(_get_Lmag)

    def _get_Ldrift(self):
        # Drift length depends on camera location
        if self.camname == 'ELANEX':
            self._Ldrift = np.float64(8.792573)
        elif self.camname == 'CMOS_FAR':
            self._Ldrift = np.float64(8.792573) + 0.8198

        return self._Ldrift
    Ldrift = property(_get_Ldrift)

    # ===================================
    # Set ypinch and eta0 to self
    # ===================================
    def _set_ypinch_eta0(self):
        """Set the ypinch and eta0 from dataset number"""
        # ===================================
        # Find and write eta_0, y_0
        # ===================================
        self.E0 = 20.35

        if self.dataset_num is None:
            logger.critical('No dataset_num detected')
            self._ypinch = np.float64(1589)
            self._eta0   = (self.Ldrift+self.Lmag/np.float64(2))*self.theta

        else:
            QS = self.setQS.energy_offset
            logger.debug('Energy offset: {}'.format(QS))

            logger.debug('Dataset_num: {}'.format(self.dataset_num))
            if self.dataset_num == '13437' or self.dataset_num == '13438':
                y0       = np.float64(1589)    # pixel position of E0 (20.35 GeV).
                eta_0_px = np.float64(949.72)  # nominal dipole dispersion in pixel, corresponding to 59.5 mm.

            elif self.dataset_num == '13448' or self.dataset_num == '13449':
                y0       = np.float64(1605.5) - np.float64(0.7923)*(self.E0+QS)  # y0 is adjusted to account for QS dispersion.
                eta_0_px = np.float64(949.72) + np.float64(0.7923)*(self.E0+QS)  # added QS dispersion of 0.7923 pix per QS GeV.

            elif self.dataset_num == '13450':
                y0       = np.float64(1655)   - np.float64(3.321)*(self.E0+QS)  # y0 is adjusted to account for QS dispersion.
                eta_0_px = np.float64(949.72) + np.float64(3.321)*(self.E0+QS)  # added QS dispersion of 3.321 pix per QS GeV.

            elif self.dataset_num == '13537':
                y0       = np.float64(1576)   + np.float(0.5193)*(20.35+QS)  # y0 is adjusted to account for QS dispersion.
                eta_0_px = np.float64(949.72) - np.float(0.5193)*(20.35+QS)  # added QS dispersion of -0.5193 pix per QS GeV.
            else:
                raise ValueError('Dataset number does not exist: {}'.format(self.dataset_num))

            if self.camname == 'ELANEX':
                # ===================================
                # Sebastien's code
                # ===================================
                z_B5D36    = np.float64(2005.65085 )  # middle of dipole magnet
                z_ELANEX   = np.float64(2015.22    )  # linac z location of ELANEX phosphor screen in meter
                z_CFAR     = np.float64(2016.04    )  # linac z location of Cherenkov Far gap in meter
                cal_ELANEX = np.float64(8.9185     )  # ELANEX camera calibration in um/pixel
                cal_CFAR   = np.float64(62.65      )  # CMOS FAR camera calibration in um/pixel

                # y0 = 259 (when QS=0) at ELANEX corresponds to y0 = 1589 on CMOS FAR.
                y0           = np.float64(259) + (cal_CFAR/cal_ELANEX) * (y0-np.float64(1589))
                self._ypinch = y0

                eta_0_px = (cal_CFAR/cal_ELANEX) * (z_ELANEX-z_B5D36) / (z_CFAR-z_B5D36) * eta_0_px

                self._eta0 = eta_0_px * cal_ELANEX * np.float64(1e-6)

            elif self.camname == 'CMOS_FAR':
                self._ypinch = y0
                self._eta0  = eta_0_px * cal_CFAR * np.float64(1e-6)
            else:
                raise ValueError('Camera name not valid: {}'.format(self.camname))

    def _get_ypinch(self):
        if self._ypinch is None:
            self._set_ypinch_eta0()
        return self._ypinch
    ypinch = property(_get_ypinch)

    def _get_eta0(self):
        self._set_ypinch_eta0()
        return self._eta0
    eta0 = property(_get_eta0)

    def energy(self, ypx):
        # ===================================
        # Assuming thin lens
        # ===================================
        if type(ypx) != np.ndarray:
            ypx = np.array([ypx])

        if self.camname == 'ELANEX':
            # For E0=20.35, ypx = 734-259 = 475, y = eta0 = 0.0548, ymotor = 0
            # This calibration comes from Sebastien and Erik
            y = (ypx)*self.res - self.ymotor + 0.050567428212613990412
        elif self.camname == 'CMOS_FAR':
            y = ypx*self.res
        else:
            raise NotImplemented('Camera not implemented: {}'.format(self.camname))

        logger.log(level=loggerlevel, msg='Eta0 is: {}'.format(self.eta0))

        approx = self.eta0*self.E0 / y
        return approx



import os
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    import re as _np
else:
    import numpy as _np


def set_QS_energy_ELANEX(E):
    VAL = (1+E/20.35)*_np.array((210.87, -164.95))  # Imaging condition for object = E200 IP or plasma exit and image = ELANEX in Fall of 2013
    # control_magnetSet({'LGPS:LI20:3261', 'LGPS:LI20:3311'}, VAL,  'action', 'TRIM');
    return VAL


def bdes2K1(bdes, E):
    bdes = _np.float64(bdes)
    E    = _np.float64(E)
    Brho = E/0.029979
    K1   = bdes/(Brho)
    return K1



import warnings
import os
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    import re as _np
    import re as h5
else:
    import numpy as _np
    import h5py as h5

import logging
logger = logging.getLogger(__name__)
loggerlevel = 9

__all__ = ['E200_create_data']


def E200_create_data(group_parent, group_name):
    # ======================================
    # Check if group name already exists
    # ======================================
    if group_name in group_parent.keys():
        logger.log(level=loggerlevel, msg='Data with name "{}" already exists.'.format(group_name))
        return group_parent[group_name]

    # ======================================
    # Create different kinds of groups
    # ======================================
    group_child = group_parent.create_group(group_name)
    group_child.create_dataset('UID', shape=(0, ), maxshape=(None, ), dtype=_np.int64)

    ref_dtype = h5.special_dtype(ref=h5.Reference)
    group_child.create_dataset('dat', shape=(0, ), maxshape=(None, ), dtype=ref_dtype)

    group_child.create_group('refs')

    return group_child



import os
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    import re as _np
    import re as h5
else:
    import numpy as _np
    import h5py as h5
import logging
logger = logging.getLogger(__name__)
from .E200_Dat import *
__all__ = ['E200_api_getdat', '_numarray2str']


def _numarray2str(numarray):
    chars = [chr(val) for val in numarray.flatten()]
    string = ''.join(chars)
    return string


def E200_api_getdat(dataset, UID=None, fieldname='dat'):
    """
    Load data from a *dataset*, which must be either an :class:`E200.Drill` or an :class:`h5py.Group` class. If no *UID* is given, all available UIDs are loaded. The *fieldname* determines which member is loaded from the *dataset*.

    Returns an instance of :class:`E200.E200_Dat`.
    """
    if type(dataset) != h5.Group:
            dataset = dataset._hdf5

    logger.log(level=10, msg='==============================')
    logger.debug('Accessing: {}'.format(dataset.name))
    # ======================================
    # Check version
    # ======================================
    version_bool = 'origin' in dataset.file.attrs.keys()
    if version_bool:
        python_version_bool = (dataset.file.attrs['origin'] == 'python-h5py')
    else:
        python_version_bool = False
    logger.debug('Matlab version: {}'.format(not python_version_bool))

    # ======================================
    # Get available UIDs in dataset
    # ======================================
    avail_uids = dataset['UID'].value

    if python_version_bool:
        if fieldname == 'dat':
            # ======================================
            # Build data from HDF5 refs
            # ======================================
            dat_refs = dataset['dat'].value
            vals = _np.empty((avail_uids.shape[0], ), dtype=_np.object)

            for i, val in enumerate(dat_refs):
                vals[i] = dataset.file[val].value

        else:
            vals = dataset[fieldname].value

    else:
        # ======================================
        # Deref if necessary
        # ======================================
        if type(dataset[fieldname][0][0]) == h5.h5r.Reference:
            vals = [dataset.file[val[0]] for val in dataset[fieldname]]
        else:
            vals = [val for val in dataset[fieldname]]

        if vals[0].shape[0] > 1:
            vals = [_np.array(val).flatten() for val in vals]
            # vals = [''.join(vec.view('S2')) for vec in vals]
            vals = [[chr(vec) for vec in val] for val in vals]
            vals = [''.join(val) for val in vals]
            vals = _np.array(vals)
        else:
            vals = [val[0] for val in vals]
            vals = _np.array(vals)

    avail_uids_num = _np.size(avail_uids)
    logger.debug('Number of available uids: {}'.format(avail_uids_num))
    if (UID is None):
        # ======================================
        # Return all results if no UID requested
        # ======================================
        out_uids = avail_uids
        out_vals = vals
    elif avail_uids_num == 1:
        # ======================================
        # Match UIDs
        # ======================================
        if avail_uids == UID:
            logger.debug('Not empty')
            out_uids = _np.array([avail_uids])
            out_vals = _np.array([vals]).flatten()
        else:
            out_uids = _np.array([])
            out_vals = _np.array([])
    else:
        # ======================================
        # Match UIDs
        # ======================================
        valbool  = _np.in1d(avail_uids, UID)
        out_vals = vals[valbool]

        out_uids = avail_uids[valbool]

    # ======================================
    # Sort UIDs
    # ======================================
    ind_sort = _np.argsort(out_uids.flatten())
    out_uids  = out_uids[ind_sort]
    out_vals = out_vals[ind_sort]

    n_uids = _np.size(out_uids)
    logger.debug('Number of UIDs found: {}'.format(n_uids))

    if n_uids > 10:
        logger.debug('Showing only first 10 UIDs')
        show_uids = out_uids[0:10]
    else:
        show_uids = out_uids

    for uid in show_uids:
        try:
            debug_uid = _np.int64(uid[0])
        except IndexError:
            debug_uid = _np.int64(uid)
        logger.debug('UID: {:d}'.format(debug_uid))
    #  out_uids = _np.array([out_uids]).flatten()

    return E200_Dat(out_vals, out_uids, field=fieldname)



import os
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
from .E200_load_images import E200_load_images
from .get_matlab import is_facet_srv
import logging
from .classes import E200_Image
logger = logging.getLogger(__name__)


# ======================================
# Image iterator
# ======================================
class E200_Image_Iter(object):
    """
    Iterable object for use in embedding in :code:`for` loops to prevent running out of memory. Loads available images from *img_dataset* (type :class:`E200.Drill` containing camera data). If *UID* is present, only loads matching UIDs. Loads *numperset* images at a time.

    *(Note: Can and should be used directly!)*
    """
    def __init__(self, img_dataset, UID=None, numperset=None):
        # ======================================
        # Check platform (only facet-srvs have
        # major memory issues)
        # ======================================
        if numperset is None:
            if is_facet_srv():
                numperset = 50
            else:
                numperset = 500

        # ======================================
        # Initialize requested iterator
        # ======================================
        self._numperset = numperset
        self._imgstr    = img_dataset
        self._uids      = UID
        self._uids_orig = UID

    def __iter__(self):
        # ======================================
        # Initialize this loop
        # ======================================
        self._subind = 0
        if self._uids is None:
            self._uids = self._imgstr.UID

        self._load_next_batch()

        return self

    def __next__(self):
        if self._imgs_ind >= self._num_uids_load:
            self._load_next_batch()

        img   = _np.array([self._images.images[self._imgs_ind]])
        dat   = _np.array([self._images.dat[self._imgs_ind]])
        uid   = _np.array([self._images.uid[self._imgs_ind]])
        imgbg = self._images.image_backgrounds
        time  = _np.array([self._images.timestamps[self._imgs_ind]])

        out = E200_Image(images=img, dat=dat, uid=uid, image_backgrounds=imgbg, timestamps=time)
        self._imgs_ind = self._imgs_ind + 1
        return out

    def _load_next_batch(self):
            # ======================================
            # Load only up to the next batch avail.
            # uids to prevent memory overflow
            # ======================================
            num_uids_left = _np.size(self._uids)
            logger.debug('Number of UIDs left: {}'.format(num_uids_left))

            if num_uids_left == 0:
                raise StopIteration

            if self._numperset == -1:
                uids = self._uids
            else:
                uid_ind = _np.min([num_uids_left, self._numperset])
                uids = self._uids[0:uid_ind]
            self._num_uids_load = _np.size(uids)

            self._images = E200_load_images(self._imgstr, UID=uids)
            self._imgs_ind = 0

            self._uids = _np.delete(self._uids, slice(0, self._num_uids_load))

            return



# import tifffile
from .E200_api_getdat import E200_api_getdat
from .classes import *  # NOQA
from .get_remoteprefix import get_remoteprefix
import PIL
import logging
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    import scipy.io as _spio
import os
loggerlevel = logging.DEBUG
logger      = logging.getLogger(__name__)


def E200_load_images(img_dataset, UID=None):
    """
    Loads available images from *img_dataset* (type :class:`E200.Drill` containing camera data) corresponding to *UID* (:code:`array` or number).

    Returns an instance of :class:`E200.E200_Image`.
    """
    logger.log(level=loggerlevel, msg='Loading images...')
    try:
        remote_bool = img_dataset._hdf5.file['data']['VersionInfo']['remotefiles']['dat'][0, 0]
    except:
        remote_bool = True
    if remote_bool:
        prefix = get_remoteprefix()
    else:
        prefix = ''

    imgdat = E200_api_getdat(img_dataset, UID=UID)

    imgs = _np.array([PIL.Image.open(os.path.join(prefix, val[1:])) for val in imgdat.dat], dtype=object)
    num_imgs = _np.size(imgs)
    timestamps = _np.empty(num_imgs)
    for i, img in enumerate(imgs):
        img_arr       = _np.asarray(img)
        imgs[i]       = _np.uint64(img_arr)
        timestamps[i] = img.tag[65002][0]

    logger.log(level=loggerlevel, msg='Loading backgrounds...')

    imgbgdat = E200_api_getdat(img_dataset, fieldname='background_dat', UID=imgdat.uid)

    # for i, val in enumerate(imgbgdat.dat):
    # print val
    i = 0
    val = imgbgdat.dat[i]
    val = os.path.join(prefix, val[1:])
    mat = _spio.loadmat(val)
    imgbg = mat['img']
        
    # if imgs[i].shape[0] == imgbg.shape[1]:
    #     imgbg = _np.transpose(imgbg)

    # imgs[i] = _np.fliplr(_np.abs(imgs[i]-_np.float64(imgbg)))

    return E200_Image(images=_np.array(imgs), dat=imgdat.dat, uid=imgdat.uid, image_backgrounds=imgbg, timestamps=timestamps)



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    from scisalt.convertH5ref import convertH5ref as _convertH5ref
from . import classes
# from convertH5ref import convertH5ref as _convertH5ref


def E200_api_getUID(dataset, val, f=None):
    """
    Return all UIDs where *val* is equal to the value of the ``dat`` member of *dataset* (which must be either an :class:`E200.Drill` or an :class:`h5py.Group` class).

    Returns an ``array`` of UIDs.
    """
    if type(dataset) == classes.Drill:
        dataset = dataset._hdf5
        
    if f is None:
        f = dataset.file
    uids = dataset['UID']
    vals = dataset['dat']

    uids = uids[:, 0]
    try:
        vals = _convertH5ref(vals, f)
    except:
        vals = _np.array(vals).flatten()

    return uids[vals == UID]



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    import re as _np
    import re as _h5
else:
    import numpy as _np
    import h5py as _h5
from .E200_api_getdat import E200_api_getdat
from .E200_Dat import *                        # noqa

__all__ = ['Data', 'Drill', 'E200_Dat', 'E200_Image']


class Data(object):
    """
    Top-level data file representing all information for a single dataset loaded from *filename*. *read_file* (type :class:`h5py._hl.files.File`, an h5py file) is the loaded data, to be read-only.

    **Should not be created directly, only accessed through** :func:`E200.E200_load_data` **or** :func:`E200.E200_load_data_gui`\ **.**
    """
    def __init__(self, read_file, filename=None):
        self._filename = filename
        self.read_file = read_file
        self.rdrill    = Drill(read_file)

        # self.data=datalevel()
        # recursivePopulate(self._data, self)

    def __enter__(self):
        return self

    @property
    def filename(self):
        """
        The *filename* the dataset was loaded with.
        """
        return self._filename

    @property
    def loadname(self):
        """
        The name of the dataset (i.e. :code:`'E217_17990'`).
        """
        return _os.path.splitext(_os.path.basename(self.filename))[0]

    def close(self):
        """
        Closes the *read_file*
        """
        self.read_file.close()

    def __exit__(self, type, value, traceback):
        self.close()


class Drill(object):
    """
    Designed for representing hierarchical h5py data. *data* (type :class:`h5py._hl.files.File`, an h5py file) is the h5py file at a particular hierarchical depth.

    Members: If not nested :class:`E200.Drill` objects, they may be arrays representing various types of information particular to their parent.
    
    **Should not be created directly, only accessed through** :func:`E200.E200_load_data` **or** :func:`E200.E200_load_data_gui`\ **.**
    """
    def __init__(self, data):
        self._hdf5 = data
        #  self._mydir = []
        for key in data.keys():
            if key[0] != '#':
                #  self._mydir.append(key)
                out = data[key]
                if type(out) == _h5._hl.group.Group:
                    setattr(self, key, Drill(out))
                elif key == 'desc':
                    desc = out.value.flatten()
                    desc = desc.view('S2')
                    desc = _np.char.decode(desc, 'UTF-8')
                    setattr(self, key, ''.join(desc))
                elif key == 'dat' and ('UID' in data.keys()):
                    uids = data['UID'].value
                    dats = E200_api_getdat(data, uids)
                    setattr(self, key, dats.dat)
                elif len(out.shape) == 2:
                    if out.shape[0] == 1 or out.shape[1] == 1:
                        out = out.value.flatten()
                    setattr(self, key, out)
                elif key == 'UID':
                    setattr(self, key, out.value)
                #  elif type(out) == _h5._hl.dataset.Dataset:
                #          if
                #          if out[0][0]==_h5.h5r.Reference:
                #                  vals=[out.file[val[0]] for val in out]
                #          else:
                #                  vals = [val for val in out]
                #          if vals[0].shape[0]>1:
                #                  vals = [np.array(val).flatten() for val in vals]
                #                  vals = [''.join(vec.view('S2')) for vec in vals]
                #                  vals = np.array(vals)
                #          else:
                #                  vals = [val[0] for val in vals]
                #                  vals = np.array(vals)
                #          setattr(self, key, vals)
                else:
                    setattr(self, key, out)

    def __repr__(self):
        out = '\<E200.E200_load_data.Drill with keys:\n_hdf5'
        for val in self._hdf5.keys():
            out = out + '\n' + val

        out = out[1:] + '\n>'
        return out


class E200_Image(E200_Dat):
    """
    Bases: :class:`E200.E200_Dat`

    Contains image data. This is designed to prevent accidental incorrect correlation of images with UIDs, in case a request for images cannot find UIDs.

    **Should not be created directly, only accessed through** :func:`E200.E200_load_images`\ **.**
    """
    def __init__(self, images, dat, uid, timestamps, image_backgrounds=None):
        self._imgs_subbed = None

        E200_Dat.__init__(self, dat, uid, field='dat')

        self._images            = images
        self._image_backgrounds = image_backgrounds
        self._timestamps        = timestamps

    @property
    def images(self):
        """
        An array of images, correlated with :attr:`E200.E200_Image.UID`.
        """
        return self._images

    @property
    def image_backgrounds(self):
        """
        An array of image backgrounds,  correlated with :attr:`E200.E200_Image.UID`.
        """
        return self._image_backgrounds

    @property
    def imgs_subbed(self):
        """
        An array of images with backgrounds subtracted,  correlated with :attr:`E200.E200_Image.UID`.
        """
        if self._imgs_subbed is None:
            self._imgs_subbed = self.images - 2*_np.fliplr(self.image_backgrounds)
        
        return self._imgs_subbed

    @property
    def timestamps(self):
        """
        An array of timestamps indicating when :attr:`E200.E200_Image.images` were taken, correlated with :attr:`E200.E200_Image.UID`.
        """
        return self._timestamps
